# Generated by Django 2.2.3 on 2019-07-29 09:08

import django.contrib.postgres.indexes
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0021_auto_20190708_0905")]

    operations = [
        migrations.RemoveIndex(
            model_name="answer", name="form_answer_value_70b2d0_idx"
        ),
        migrations.AddIndex(
            model_name="answer",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["meta", "value"], name="form_answer_meta_98edaa_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="document",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["meta"], name="form_docume_meta_333f0b_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="form",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["meta"], name="form_form_meta_a2eed1_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="option",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["meta"], name="form_option_meta_c09f31_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="question",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["meta"], name="form_questi_meta_ea1ba0_gin"
            ),
        ),
    ]
